import csv
from progressbar import ProgressBar

from django.core.management.base import BaseCommand, CommandError
from django.contrib.gis.geos import Point

from ...models import Place

# Taken from the tech spec http://www.ordnancesurvey.co.uk/oswebsite/docs/technical-specifications/50k-gazetteer-technical-specification.pdf
FIELDNAMES = (
    "SEQ",         # Unique sequence number of the record.
    "KM_REF",      # The National Grid 1 km by 1 km square the feature or centre of the feature falls within on the 1:50 000 scale mapping.
    "DEF_NAM",     # Distinctive name of the feature the record describes
    "TILE_REF",    # The 1:50 000 Scale Colour Raster tile the feature or centre of the feature falls within.

    # The latitude and longitude degrees and minutes of the 1 km National Grid square the feature or centre of the feature falls within.
    "LAT_DEG", "LAT_MIN", "LONG_DEG", "LONG_MIN",

    # National Grid position of the centre point of the 1 km square the feature or centre of the feature falls within. 352500
    "NORTH", "EAST",

    "GMT",         # Greenwich Mean Time Position in relation to the Greenwich Meridian.
    "CO_CODE",     # Code representing the county or unitary authority boundary the definitive name falls within. See annexe B for a list.
    "COUNTY",      # Abbreviated county or unitary authority name.
    "FULL_COUNTY", # Full county or unitary authority name.
    "F_CODE",      # Describes what the feature is. See chapter 6 for complete list of definitions. 'T' represents town.
    "E_DATE",      # The date the record was last amended.
    "UPDATE_CO",   # Update code 'I' represents insert. 'A' represents amendment. 'D' represents deletion.
    "SHEET_1",     # Primary sheet no
    "SHEET_2",     # Second sheet no
    "SHEET_3",     # Third sheet no
)

class Command(BaseCommand):
    help = 'Load places from the OS 1:50k Gazetteer database'

    def handle(self, *args, **options):
        filename = args[0]
        reader = csv.DictReader(open(filename), fieldnames=FIELDNAMES, delimiter=':', quotechar='"')
        rownum = 0

        # hacky way to approximate the number of rows in the file
        pbar = ProgressBar( maxval=len(list(open(filename))) ).start()

        # There is nothing reliabe to use for detecting duplicates, delete database instead.
        Place.objects.filter(source=Place.SOURCE_OS_50K_GAZETEER).delete()

        for row in reader:
            rownum += 1
            pbar.update(rownum)

            name     = row['DEF_NAM']
            county = row['FULL_COUNTY']
            osgb36_x = int(row['EAST'])
            osgb36_y = int(row['NORTH'])

            # Skip blank lines, and roads which no name (like "A1")
            if not name:
                continue

            # create the name
            context_name = "{0}, {1}".format(name, county)

            # create a point for the centre
            centre = Point(osgb36_x, osgb36_y, srid=Place.SRID_OSGB36)
            centre.transform(Place.SRID_WGS84)

            # create the place
            place = Place(
                name = name,
                context_name = context_name,
                centre = centre,
                source = Place.SOURCE_OS_50K_GAZETEER,
            )

            # skip entries outside the allowed bonding boxes
            if not place.is_in_allowed_bounding_boxes():
                continue

            place.save()

        pbar.finish()



import csv
from progressbar import ProgressBar

from django.core.management.base import BaseCommand, CommandError
from django.contrib.gis.geos import Point

from ...models import Place


# Taken from the tech spec http://www.ordnancesurvey.co.uk/oswebsite/docs/user-guides/os-locator-user-guide.pdf
FIELDNAMES = (
    "Name", # Feature name
    "Classification", # Road Number Classification
    "Centx", # X coord for Centre point of road object
    "Centy", # Y coord for Centre point of road object
    "Minx",  # X coord for SW corner of the road object box
    "Maxx",  # X cord for NE corner of road object box
    "Miny",  # Y coord for SW corner of the road object box
    "Maxy",  # Y coord for NE corner of road object box
    "Settlement", # Town in which the centre of the object falls
    "Locality", # Dataset with a description for the point at the centre of the object
    "Cou_Unit", # County or Unitary Authority in which the centre of the object falls, according to the latest Boundary-Line dataset
    "Local Authority", # Local Authority in which the centre of the object falls, according to the latest Boundary-Line dataset
    "Tile_10k", # 1:10 000 scale tile reference for the centre point of the object
    "Tile_25k", # 1:25 000 scale tile reference for the centre point of the object
    "Source", # Source of information
)

class Command(BaseCommand):
    help = 'Load places from the OS Locator database'

    def handle(self, *args, **options):
        filename = args[0]
        reader = csv.DictReader(open(filename), fieldnames=FIELDNAMES, delimiter=':', quotechar='"')
        rownum = 0

        # hacky way to approximate the number of rows in the file
        pbar = ProgressBar( maxval=len(list(open(filename))) ).start()

        # There is nothing reliabe to use for detecting duplicates, delete database instead.
        Place.objects.filter(source=Place.SOURCE_OS_LOCATOR).delete()

        for row in reader:
            rownum += 1
            pbar.update(rownum)

            name     = row['Name']
            locality = row['Locality']
            osgb36_x = int(row['Centx'])
            osgb36_y = int(row['Centy'])

            # Skip blank lines, and roads which no name (like "A1")
            if not name:
                continue

            # create the name
            context_name = "{0}, {1}".format(name, locality)

            # create a point for the centre
            centre = Point(osgb36_x, osgb36_y, srid=Place.SRID_OSGB36)
            centre.transform(Place.SRID_WGS84)

            # create the place
            place = Place(
                name = name,
                context_name = context_name,
                centre = centre,
                source = Place.SOURCE_OS_LOCATOR,
            )

            # skip entries outside the allowed bonding boxes
            if not place.is_in_allowed_bounding_boxes():
                continue

            place.save()

        pbar.finish()




from django.test import TestCase
from django.test.utils import override_settings

from ..models import Place

from django.contrib.gis.geos import Point


class GeocoderModelTests(TestCase):

    @override_settings(GEOCODER_BOUNDING_BOXES=(
        # xmin, ymin, xmax, ymax
        ( -1,   50,   1,    52 ),
    ))
    def test_is_in_allowed_bounding_boxes(self):
        defaults = dict(context_name="Context", source=Place.SOURCE_OS_LOCATOR)
        good_place = Place(name='Allowed', centre=Point(0, 51),  **defaults)
        bad_place  = Place(name='Bad',     centre=Point(-2, 51), **defaults)

        self.assertTrue(good_place.is_in_allowed_bounding_boxes())
        self.assertFalse(bad_place.is_in_allowed_bounding_boxes())



from django.test import TestCase
from django.core.urlresolvers import reverse
from django.conf import settings
from django.contrib.gis.geos import Point
from django.utils import simplejson as json

from ..models import Place

class GeocoderViewTests(TestCase):

    def setUp(self):

        self.lookup_url = reverse('geocoder-lookup')

        names = [
            'Bedford Place',
            'Bevan Avenue',
            'Bevan Close',
            'Bevan Way',
            'Chelsea',
        ]

        centre = Point(50, 2)

        for name in names:
            Place.objects.create(
                name = name,
                context_name = name + ", London",
                centre = centre,
                source = Place.SOURCE_OS_LOCATOR,
            )

    def get_lookup(self, term):
        resp = self.client.get(self.lookup_url, {"term": term})
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp['Content-Type'], 'application/json')
        all = json.loads(resp.content)
        texts = [ x['text'] for x in all ]
        return { "all": all, "texts": texts }


    def test_no_matches(self):
        self.assertEqual( self.get_lookup('This should not match anything')['all'], [] )

    def test_one_matches(self):
        self.assertEqual( self.get_lookup('chelsea')['texts'], ['Chelsea, London'] )

    def test_many_matches(self):
        self.assertEqual(
            self.get_lookup('Bevan')['texts'],
            ['Bevan Avenue, London', 'Bevan Close, London', 'Bevan Way, London']
        )





from django.db import models
from django.contrib.gis.db import models as geomodels
from django.contrib.gis.geos import Polygon

from django.conf import settings

class Place(geomodels.Model):

    # Required for coordinate conversions
    SRID_WGS84  = 4326
    SRID_OSGB36 = 27700

    # Name is shorter and should be used when searching.
    name = models.CharField(max_length=250, db_index=True)

    # Longer name, eg: "High Street, Putney" that can be displayed to user.
    context_name = models.CharField(max_length=250)

    # Where is the center of this place. Might not be very accurate.
    centre = geomodels.PointField()

    # Where did this entry come from?
    SOURCE_OS_LOCATOR = 1
    SOURCE_OS_50K_GAZETEER = 2
    SOURCE_CHOICES = (
        (SOURCE_OS_LOCATOR, 'OS Locator'),
        (SOURCE_OS_50K_GAZETEER, 'OS 1:50k Gazetteer'),
    )
    source = models.IntegerField(choices=SOURCE_CHOICES)

    objects = geomodels.GeoManager()


    def is_in_allowed_bounding_boxes(self):
        for bbox in settings.GEOCODER_BOUNDING_BOXES:
            polygon = Polygon.from_bbox(bbox)
            if polygon.contains(self.centre):
                return True
        return False



import sys
import os

from django.test import TestCase
from django.test.utils import override_settings
from django.core.management import call_command
from django.forms.models import model_to_dict

from ..models import Place


class DevNull(object):
    def write(self, data):
        pass


class GeocoderCsvImportTests(TestCase):

    def setUp(self):

        # Mute stderr (the progress bar output)
        self.old_stderr = sys.stderr
        sys.stderr = DevNull()

        # Paths to the various sample data files
        csv_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_data')
        self.os_locator_data_filename       = os.path.join(csv_dir, 'OS_Locator2013_1_OPEN_sample.txt')
        self.os_50k_gazetteer_data_filename = os.path.join(csv_dir, '50kgaz2013_sample.txt')

    def tearDown(self):
        sys.stderr = self.old_stderr

    @override_settings(GEOCODER_BOUNDING_BOXES=(
        # xmin, ymin, xmax, ymax
        ( -0.8, 51.1, 0.6,  51.8 ), # London
    ))
    def test_os_locator(self):

        call_command('load_places_from_OS_locator', self.os_locator_data_filename)
        self.assertEqual(Place.objects.count(), 10)

        place = Place.objects.get(name="BEVAN COURT")
        self.assertEqual(
            model_to_dict(place, exclude=['centre','id']),
            {
                'name': 'BEVAN COURT',
                'context_name': 'BEVAN COURT, Waddon',
                'source': Place.SOURCE_OS_LOCATOR,
            }
        )
        # We have to be _almost_ equal hee because different geo packages are more or
        # less accurate in their conversion of eastings/northings to lat/lon
        self.assertAlmostEqual(place.centre.x, -0.11587620309898315, places=3)
        self.assertAlmostEqual(place.centre.y, 51.361488927177462, places=3)

    @override_settings(GEOCODER_BOUNDING_BOXES=(
        # xmin, ymin, xmax, ymax
        ( -0.8, 51.1, 0.6,  51.8 ), # London
    ))
    def test_50k_gazetteer(self):

        call_command('load_places_from_OS_50k_gazetteer', self.os_50k_gazetteer_data_filename)
        self.assertEqual(Place.objects.count(), 10)

        place = Place.objects.get(name="Farringdon Station")
        self.assertEqual(
            model_to_dict(place, exclude=['centre','id']),
            {
                'name': 'Farringdon Station',
                'context_name': 'Farringdon Station, City of London',
                'source': Place.SOURCE_OS_50K_GAZETEER,
            }
        )
        # We have to be _almost_ equal hee because different geo packages are more or
        # less accurate in their conversion of eastings/northings to lat/lon
        self.assertAlmostEqual(place.centre.x, -0.1061755458136722, places=3)
        self.assertAlmostEqual(place.centre.y, 51.517070389322676, places=3)



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding index on 'Place', fields ['name']
        db.create_index('geocoder_place', ['name'])


    def backwards(self, orm):
        # Removing index on 'Place', fields ['name']
        db.delete_index('geocoder_place', ['name'])


    models = {
        'geocoder.place': {
            'Meta': {'object_name': 'Place'},
            'centre': ('django.contrib.gis.db.models.fields.PointField', [], {}),
            'context_name': ('django.db.models.fields.CharField', [], {'max_length': '250'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '250', 'db_index': 'True'}),
            'source': ('django.db.models.fields.IntegerField', [], {})
        }
    }

    complete_apps = ['geocoder']


from django.views.generic import ListView

from django.http import HttpResponse
from django.utils import simplejson as json

from .models import Place


class Lookup(ListView):
    model = Place

    def get_queryset(self):
        qs = super(Lookup, self).get_queryset()

        term = self.request.GET.get('term', '')
        qs = qs.filter(name__istartswith=term)
        qs = qs.order_by('name')

        return qs

    def render_to_response(self, context, **kwargs):

        to_serialize = []
        for obj in context['object_list'][:8]:
            to_serialize.append({
                "id":   obj.id,
                "text": obj.context_name,
                "lat":  obj.centre.y,
                "lon":  obj.centre.x,
            });

        json_string = json.dumps( to_serialize, sort_keys=True, indent=4 )

        kwargs['content_type'] = 'application/json'
        return HttpResponse(json_string, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import os

file_dir = os.path.abspath(os.path.dirname(__file__))

def read_file(filename):
    filepath = os.path.join(file_dir, filename)
    return open(filepath).read()

setup(
    name="django-os-geocoder",
    version='0.1.6',
    description='A Django app that lets you load Ordnance Survey data and use it to find places.',
    long_description=read_file('README.rst'),
    author='mySociety',
    author_email='programmers@mysociety.org',
    url='https://github.com/mysociety/django-os-geocoder',
    packages=find_packages(),
    include_package_data=True,
    install_requires=read_file('requirements.txt'),
    classifiers=[
        'Framework :: Django',
    ],
)
